#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Game of Life console. 
small play ground for leanning python, or just having fun.

To use the console with an empty grid: 
    python -i -mgof.console
or
    bpython -i -mgof.console

To use the console with a pseudo animation: 
    python -i -mgof.demo

Avalailble variable : 
* grid your playground a matrix of cellular automata
* patterns : pixel, still, oscillator, glider
    they are singular patterns to play with in game of life.
* all_pattern : a list of all patterns (except pixel)
* matrix : the class name of grid is imported for educationnal purpose
* DEAD, ALIVE

Available functions:
* intro() : a short summary of all available functions
* bleach(...) a function to init the grid
* at(...) a function to draw a pattern on the grid
* rand_pattern() : a function to add random pattern in your grid
* evolve(...) make the game evolve for some time. If your terminal and/or 
	interactive python supports it, it will make a continuous animation

* use help(function_name) to know more  yes it is a builtin ^_^

"""
from matrix import matrix
from time import sleep
from .gof import glider, oscillator, still, pixel, all_pattern
from .gof import evolve, bleach,dirty, DEAD, ALIVE, at
#### Constants and globals
__all__ = [ 
    "matrix","at",
    "grid", "intro", 
    "glider", "oscillator", "still","pixel","all_pattern",
    "evolve", "bleach", "dirty", "DEAD", "ALIVE" ]

x=10
y=30
def intro(): 
    print __doc__
    print """
    you are left with an empty grid of %dx%d to play with, have fun""" % (x,y)

grid=matrix(x,y,x*y*[DEAD])



if '__main__' == __name__:
    print __doc__




#!/usr/bin/python
import os
from .console import *
x,y=16,14
speed=5
### All these works ! 
#grid = matrix(x, y, bytearray(x*y))
#from numpy import array,zeros
#grid = matrix(x, y, zeros(x*y))
#grid = matrix(x, y, [DEAD]* x*y)
#from collections import defaultdict
#grid = matrix(x,y,defaultdict(int,{}))
from .weird_array import Bitmap, SparseArray
grid = matrix(x,y,Bitmap(ALIVE<<x*y))
#grid = matrix(x,y,SparseArray(set()))

at(grid, 3,9, glider)
at(grid, 2,2, oscillator)
at(grid, 6,5, still)

def reset(grid=grid):
    grid.__init__(x,y,Bitmap(ALIVE<<x*y))

evolve(grid, 160,speed)




#!/usr/bin/python
# -*- coding: utf-8 -*-
from random import randint
"""a 2D accessor for 1 dimensional array, which is backend agnostic.
you need a MutableSequence as a constructor
"""
def ift(p, vt, vf):
    return lambda val: (ift(*vt)(val) if isinstance(vt, tuple) else vt) \
        if p(val) else (ift(*vf)(val) if isinstance(vf, tuple) else vf)

compile_nbit = lambda nbit: lambda f: sum([int(f(n))<<n for n in range(1<<nbit)])
compile = compile_nbit(9)
hamming = lambda x: sum([ (x>>i)&1 for i in range(x.bit_length()) ])
interpret = lambda code: lambda val: int((1<<(val))&code != 0)
conway_prog= ift(
    lambda x: (x&1) == 1,
        (lambda y:3 <= hamming(y) <= 4, 1, 0),
        (lambda z:hamming(z)       ==3, 1, 0),
)
assert conway_prog(7)

conway_code = compile(conway_prog)
conway = interpret(conway_code)
assert hamming(3) == 2
assert hamming(1) == 1
assert hamming(7) == 3
assert hamming(0b01111) == 4
assert conway(0b00001) == 0
assert conway_code
assert conway(0b01111) == 1
assert conway(0b11111) == 0
assert conway(0b11110) == 0
assert conway(0b11100) == 1
assert conway(0b10101) == 1

class matrix:

    def __init__(self,size_x,size_y, mutable_sequence):
        """construct a view of a dimension x and y on mutable_sequence
        the mutable_sequence must have a dimension compliant with its size
        """
        self.size_y=size_y
        self.size_x=size_x
        self._compute = conway
        self._code = conway_code
        self.matrix= mutable_sequence
        self.pattern = [ ".", "X" ]

    def mutate(self, nb_flip=1):
        for i in range(nb_flip):
            self._code = self._code^(1<<randint(0,self._code.bit_length()-1))
        self._compute=interpret(self._code)

    def get_rand(self):
        from random import randint 
        return self.matrix[ randint(0,self.size_y * self.size_x -1 ) ]

    def _oneD_offset(self,ix,iy):
        
        x=ix%self.size_x
        y=iy%self.size_y
        offset = y*self.size_x+x
        if offset >= self.size_x * self.size_y :
            print "%d(%d), %d(%d) => %d BOOOM"% (x,ix, y,iy, offset)
        return offset
        
    def get(self,x,y):
        """get item at coordinates x,y"""
        return self.matrix[self._oneD_offset(x,y)]
   
    def map_neighbor_to_int(self,x,y):
        """mis placed method
        TODO correct it one day"""
        # LSB counting of neighbors (self is bit 0 , north bit 1 .....)
        ordered_neighbors= [    (0,0), 
            (0,1),(-1,1),(-1,0),(-1,-1), (0,-1),(1,-1), (1,0), (1,1)]

        return sum([
            int(not(not(self.get(x+dx,y+dy))))<<i for i,(dx, dy) in enumerate(ordered_neighbors)
        ])
    def compute_state(self,x, y):
        return self._compute(self.map_neighbor_to_int(x,y))

    def set(self,x,y,val):
        """set value val at coordinates x, y"""
        self.matrix[self._oneD_offset(x,y)]=val
    
    def duplicate(self):
        """return a copy of itself. Maybe copy would be a better name ? """
        
        copied_matrix = None
        if hasattr(self.matrix, "copy"):
            copied_matrix = self.matrix.copy()
        else:
            copied_matrix = [ x for x in self.matrix ] 
        return matrix(self.size_x, self.size_y,copied_matrix )

    def copy_in(self,other):
        """copy the underlying mutable_sequence in an other matrix"""
        if hasattr(self.matrix, "copy"):
            copied_matrix = self.matrix.copy()
        else:
            copied_matrix = [ x for x in self.matrix ] 
        other.matrix=copied_matrix

        
    
    def __str__(self):
        """poor man's amazing graphical effects:)"""
        to_print=" "
        to_print+="'    " * (self.size_y/5 )
        for x in range(self.size_x):
            for y in range(self.size_y):
                if (y==0):
                    to_print+='\n ' if x%5 else "\n-"
                to_print+="%1s" % ( self.pattern[self.get(x,y)] )
        return to_print

def matrix_check():
#matrice 2x2
    m=matrix(3,2, [0]*6)
    # haut à gauche
    m.set(0,0,1)
    # haut mil
    m.set(1,0,2)
    m.set(2,0,3)
    #mil bas
    m.set(1,1,7)
    print(m)
    print(bin(m.map_neighbor_to_int(0,1)))
    print(bin(m.map_neighbor_to_int(0,0)))
    print(u""" 
should be: 
     0  1
 0   X  .
 1   X  X
 2   X  .

""")
if __name__ == '__main__':
    matrix_check()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""facade design pattern to make : set() and int look like a mutable sequence
of 0 and 1"""

class SparseArray(object):
    """a mtuable sequence facade for set()"""
    def __init__(self,_set=None):
        self._set = _set if _set else {0}^{0}

    def __setitem__(self,offset, val):
        if val:
            self._set.add(offset)
        else:
            self._set.discard(offset)

    def __getitem__(self,offset):
        return int(bool(offset in self._set))

    def copy(self):
        return SparseArray(self._set.copy())


class Bitmap(object):
    """bit array view on an int"""
    def __init__(self,_int=None):
        if not _int:
            raise Exception("Please provide a non null int")

        self._int = _int

    def __setitem__(self,offset, val):
        if val:
            self._int |= 1<<offset
        else:
            self._int &= ~(1<<offset)

    def __getitem__(self,offset):
        return int(bool(self._int & 1<<offset))

    def copy(self):
        return Bitmap(self._int)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from random import choice, randint
import os
from time import sleep
oscillator = [(0,0),(0,1),(0,2)]
still = [(0,0),(0,1),(1,0),(1,1)]
glider = [(0,2),(1,2),(2,2),(2,1),(0,0)]
pixel = [(0,0)]

DEAD=0
ALIVE=1

all_pattern = [ oscillator, still, glider ]

##### helpers

def at(grid,x,y, pattern, value=ALIVE):
    """transpose a pattern at position X,Y in a matrix object with value ALIVE 
    by default to draw a single point at coordonates x=2, y=3 use 
    >>> at(grid 2,3 , pixel)

    """
    for dx,dy in pattern:
        grid.set(x+dx,y+dy,value)

def dirty(grid, times=1, all_pattern=all_pattern):
    """add random patterns n times at random position in a matrix object"""


    while times:
        times-=1
        at( grid, 
            randint(0, grid.size_x) ,
            randint(0,grid.size_y),
            choice(all_pattern)
        )

def evolve(grid,times=1, speed=1.0):
    """make grid evolve for times turn. Print all the grids if on linux 
    at speed times the default. 
    Cheat code : if speed is "unseeable" then the grid evolve without printing
    """

    if not hasattr(grid,'time'):
        grid.time=0
    time = grid.time
    
    n_grid = grid.duplicate()
    while times:
        times-=1
        time+=1
        if 'unseeable' != speed:
            os.system('clear')
            print grid
            print "time:<%3d>" % time
            sleep(1.0/speed)
        
        for x in range(grid.size_x):
            for y in range(grid.size_y):
                n_grid.set(x, y, grid.compute_state(x,y))
        n_grid.copy_in(grid)


    grid.time=time
    


def bleach(grid, x,y,empty_matrix=None):
    """time to bleach the grid and remove all leaving cells for a new start!
    The thrid parameter is an easter eggs to use any matrix object backend
    try  : 
    >>> from collections import defaultdict
    >>> bleach(grid, 10,10, defaultdict(bool, {}))
    
    to change the default implementation based on a boring array to a 
    defaultdict. 
    or 
    
    >>> from gof.weird_array import Bitmap
    >>> bleach(grid, 10,10, 1<<(10*10))
    

    """
    if not empty_matrix:
        empty_matrix=x*y*[DEAD]
    grid.__init__(x,y, empty_matrix)
    grid.time=0




#!/usr/bin/env pyrhon
# -*- coding: "utf-8" -*-

from distutils.command.build_py import build_py as _build_py
from distutils.core import setup

def test():
    """Specialized Python source builder."""
setup(
        name='gof',
        version='0.1.2',
        author='Julien Tayon',
        author_email='julien@tayon.net',
        packages=['gof'],
        url='http://gof.readthedocs.org/',
        license='LICENSE.txt',
        description='''a little fun with Conway's Game Of Life''',
        requires=[ ],
        classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: Python Software Foundation License',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: POSIX',
          'Programming Language :: Python',
          ],
)



#!/usr/bin/python
import os
from .console import *
from time import sleep
from .matrix import hamming
x,y=16,14
speed=5
### All these works ! 
#grid = matrix(x, y, bytearray(x*y))
#from numpy import array,zeros
#grid = matrix(x, y, zeros(x*y))
#from collections import defaultdict
#grid = matrix(x,y,defaultdict(int,{}))
from .weird_array import Bitmap, SparseArray
grid = matrix(x,y,Bitmap(ALIVE<<x*y))
grid_p = matrix(x,y,Bitmap(ALIVE<<x*y))
#grid = matrix(x,y,SparseArray(set()))

at(grid, 3,9, glider)
at(grid, 2,2, oscillator)
at(grid, 6,5, still)
dirty(grid, 5)

grid_p.matrix._int = grid.matrix._int

grid_p.mutate(5)

grid_f = matrix(x, y, [DEAD]* x*y)
grid_f.pattern = [ " ", "x" , "." , "X" ]
print grid_f
for frame in range(200):
    evolve(grid, 1,"unseeable")
    evolve(grid_p, 1,"unseeable")

    for l in range(grid.matrix._int.bit_length()-1):
        grid_f.matrix[l] = grid.matrix[l] + 2 * grid_p.matrix[l]
    os.system('clear')

    print "\n".join(map(lambda t:"%s\t%s\t%s" %(t),zip(str(grid).split("\n"),str(grid_p).split("\n"),str(grid_f).split("\n"))))
    print "time is %d" % grid.time
    print "pertubation %d"  % (hamming(grid.matrix._int^grid_p.matrix._int))
    sleep(.2)
